/* Copyright 2010 Jetro Lauha - http://iki.fi/jetro/
 *
 * $Id: MainState.cpp 242 2010-12-20 02:01:25Z tonic $
 * $Revision: 242 $
 */

#include <time.h>
#include <assert.h>
#include <algorithm>

#include "main.h"
#include "MainState.h"


#define PLR_SPEED   0.075f
#define PLR_W       0.6f
#define PLR_H       0.6f

#define FRAMES_PER_DIR  2
#define TIME_PER_FRAME  250

#define DIG_PICKAXE_TIME        500
#define DIG_DRILL_TIME          250
#define DIG_PICKAXE_HIT         0.125f
#define DIG_DRILL_HIT           2 * DIG_PICKAXE_HIT
#define GROUND_KILL_THRESHOLD   (1 - 3 * DIG_PICKAXE_HIT + 0.001f)
#define GROUND_KILL_SPEED       0.015f

#define SCANNER_MIN_ANGLE       ((float)PI / 4)
#define SCANNER_MAX_ANGLE       ((float)PI)
#define SCANNER_MIN_DIST        6.0f
#define SCANNER_MAX_DIST        12.0f
#define SCANNER_DIST_CHANGE     0.01f
#define SCANNER_ANGLE_CHANGE    0.05f
#define SCANNER_TIME            1000

#define GOAL_DIAMONDS           5
#define DIAMOND_GOLD_VALUE      10000
#define START_GOLD              2000

#define PICKAXE_USE_COST        1
#define DRILL_COST              200
#define DRILL_COST_STR          "$200"
#define DRILL_USES_MIN          50
#define DRILL_USES_MAX          170

#define SCANBAT_COST            100
#define SCANBAT_COST_STR        "$100"


static const char *sItemNames[ITEM_COUNT] =
{
    "Pickaxe",
    "Drill",
    "Scanner"
};

enum SAMPLE_ID
{
    SAMPLE_PICKAXE1 = 0,
    SAMPLE_PICKAXE2,
    SAMPLE_PICKAXE3,
    SAMPLE_PICKAXE4,
    SAMPLE_PICKAXE5,
    SAMPLE_DRILL1,
    SAMPLE_DRILL2,
    SAMPLE_DRILL3,
    SAMPLE_EXPL1,
    SAMPLE_EXPL2,
    SAMPLE_EXPL3,
    SAMPLE_COLLECT,
    SAMPLE_SCANNER,
    SAMPLE_COUNT, //

    SAMPLE_PICKAXE_FIRST = SAMPLE_PICKAXE1,
    SAMPLE_PICKAXE_COUNT = 5,
    SAMPLE_DRILL_FIRST = SAMPLE_DRILL1,
    SAMPLE_DRILL_COUNT = 3,
    SAMPLE_EXPL_FIRST = SAMPLE_EXPL1,
    SAMPLE_EXPL_COUNT = 3
};

static char *sSampleFileNames[SAMPLE_COUNT] =
{
    "data/pickaxe1.ogg",
    "data/pickaxe2.ogg",
    "data/pickaxe3.ogg",
    "data/pickaxe4.ogg",
    "data/pickaxe5.ogg",
    "data/drill1.ogg",
    "data/drill2.ogg",
    "data/drill3.ogg",
    "data/expl1.ogg",
    "data/expl2.ogg",
    "data/expl3.ogg",
    "data/collect1.ogg",
    "data/scanner.ogg"
};

#define AUDIO_CHANNELS 32

static FMUSIC_MODULE *sMusic;
static FSOUND_SAMPLE *sSamples[SAMPLE_COUNT];


MainState::MainState()
{
    mWorld = 0;
    mKeyUp = mKeyDown = mKeyLeft = mKeyRight = false;
}


MainState::~MainState()
{
}


static bool sStringStartsWith(const c8 *str, const c8 *beginning)
{
    irr::core::stringc s(str), b(beginning);
    return s.subString(0, b.size()) == b;
}


bool MainState::init()
{
    Graphics *g = App::getInstance()->getGraphics();

    if (!FSOUND_Init(44100, AUDIO_CHANNELS, 0))
    {
        gErrorMessage("Audio error", "Unable to initialize sound! (%d)",
                      FSOUND_GetError());
    }
    for (u32 a = 0; a < SAMPLE_COUNT; ++a)
    {
        sSamples[a] = FSOUND_Sample_Load(FSOUND_FREE, sSampleFileNames[a], FSOUND_LOOP_OFF, 0, 0);
        if (!sSamples[a])
            gErrorMessage("Audio error", "Can't load %s", sSampleFileNames[a]);
    }
    sMusic = FMUSIC_LoadSong("data/ld48-19.oxm");
    if (!sMusic)
        gErrorMessage("Audio error", "Can't load music");
    FMUSIC_PlaySong(sMusic);
    FMUSIC_SetMasterVolume(sMusic, 128);

    mFont = new BMFont(g, "data/droid30.fnt");
    if (!mFont->init())
        return false;

    // Make sure these are in synch with IMAGE_ID
    mImages.push_back(g->loadImage("data/button.png"));
    mImages.push_back(g->loadImage("data/button-hot.png"));
    mImages.push_back(g->loadImage("data/button-active.png"));
    mImages.push_back(g->loadImage("data/button-hot-active.png"));
    mImages.push_back(g->loadImage("data/sprites.png"));
    mImages.push_back(g->loadImage("data/cave.png"));
    mImages.push_back(g->loadImage("data/scanner.png"));

    mButtonStyle = new IMGUIImageButtonStyle(mFont, mImages[IMG_BUTTON], mImages[IMG_BUTTON_HOT], mImages[IMG_BUTTON_ACTIVE], mImages[IMG_BUTTON_HOT_ACTIVE]);

    mLabelText = "";

    mDevice = App::getInstance()->getIrrDevice();
    mVideo = App::getInstance()->getIrrVideoDriver();

    mLastMousePos.set(WIDTH / 2, HEIGHT / 2);

    mTitleUIActive = true;
    mInGame = false;
    //startGame(); // testing

    return mFont != NULL && mButtonStyle != NULL;
}


void MainState::deinit()
{
    UI32 a;
    delete mFont;
    for (a = 0; a < mImages.size(); ++a)
        delete mImages[a];
    mImages.clear();

    FMUSIC_StopAllSongs();
    FMUSIC_FreeSong(sMusic);
    for (a = 0; a < AUDIO_CHANNELS; ++a)
        FSOUND_StopSound(a);
    for (a = 0; a < SAMPLE_COUNT; ++a)
        FSOUND_Sample_Free(sSamples[a]);
    FSOUND_Close();

    delete[] mWorld;
}


void MainState::onActivate(UI32 time)
{
    AppState::onActivate();
    mPaused = false;
}


void MainState::startGame()
{
    const int openAirHeight = 10;

    delete[] mWorld;
    mLevelWidth = 30;
    mLevelHeight = 50;
    mWorld = new Tile[mLevelWidth * mLevelHeight];

    for (int y = 0; y < openAirHeight; ++y)
    {
        for (int x = 0; x < mLevelWidth; ++x)
            mWorld[y * mLevelWidth + x].type = TT_OPEN_AIR;
    }
    for (int x = 0; x < mLevelWidth; ++x)
        mWorld[openAirHeight * mLevelWidth + x].type = TT_IMPASSABLE;
    for (int y = openAirHeight; y < mLevelHeight; ++y)
        mWorld[y * mLevelWidth].type = mWorld[y * mLevelWidth + mLevelWidth - 1].type = TT_IMPASSABLE;
    mWorld[openAirHeight * mLevelWidth + mLevelWidth / 2].type = TT_TUNNEL;
    mWorld[(openAirHeight + 1) * mLevelWidth + mLevelWidth / 2].type = TT_TUNNEL;

    mWorld[(openAirHeight - 2) * mLevelWidth + mLevelWidth / 2 - 2].type = TT_SHOP_DRILL;
    mWorld[(openAirHeight - 2) * mLevelWidth + mLevelWidth / 2 + 2].type = TT_SHOP_SCANBAT;

    srand((unsigned int)time(NULL));
    for (int a = 0; a < GOAL_DIAMONDS; ++a)
    {
        bool found = false;
        while (!found)
        {
            int tx = rand() % (mLevelWidth - 2) + 1;
            int ty = rand() % (mLevelHeight - openAirHeight - 2) + openAirHeight + 2;
            Tile &t = mWorld[ty * mLevelWidth + tx];
            if (t.type == TT_GROUND && !t.hasGold)
            {
                t.hasGold = true;
                found = true;
            }
        }
    }

    //mWorld[(openAirHeight + 3) * mLevelWidth + mLevelWidth / 2].hasGold = true; // Testing

    mScore = START_GOLD;
    mPlayer.reset();
    mPlayer.state = PLR_MOVING;
    mPlayer.x = mLevelWidth / 2.0f;
    mPlayer.y = openAirHeight - 0.5f;

    mScanResultTransition = 100.f;

    mTitleUIActive = false;//true;
    mInGame = true;//false;
}


bool MainState::onKeyUp(const irr::SEvent::SKeyInput &event)
{
    if (event.Key == KEY_LEFT)
        mKeyLeft = false;
    else if (event.Key == KEY_RIGHT)
        mKeyRight = false;
    else if (event.Key == KEY_UP)
        mKeyUp = false;
    else if (event.Key == KEY_DOWN)
        mKeyDown = false;

    return false;
}


static float smod(float x, float m)
{
  return x - ((floorf(x / m + 0.5f)) * m);
}

static float unwrap(float prev, float next, float range)
{
  return prev + smod(next - prev, range);
}


static bool angleBetween(float n, float a, float b)
{
    n = fmodf(fmodf(n, (float)(2 * PI)) + (float)(2 * PI), (float)(2 * PI));
    a = fmodf(fmodf(a, (float)(2 * PI)) + (float)(2 * PI), (float)(2 * PI));
    b = fmodf(fmodf(b, (float)(2 * PI)) + (float)(2 * PI), (float)(2 * PI));
    if (a < b)
        return a <= n && n <= b;
    return (0 <= n && n <= b) || (a <= n && n <= (float)(2 * PI));
}


bool MainState::onKeyDown(const irr::SEvent::SKeyInput &event)
{
    if (event.Key == KEY_ESCAPE)
        App::getInstance()->setAlive(false);

    if (event.Key == KEY_LEFT)
        mKeyLeft = true;
    else if (event.Key == KEY_RIGHT)
        mKeyRight = true;
    else if (event.Key == KEY_UP)
        mKeyUp = true;
    else if (event.Key == KEY_DOWN)
        mKeyDown = true;

    if (event.Key == KEY_KEY_X && mPlayer.state == PLR_MOVING)
    {
        mPlayer.selItem = (mPlayer.selItem + 1) % ITEM_COUNT;

        if (mPlayer.selItem == ITEM_DRILL && mPlayer.drillUses == 0)
            mPlayer.selItem = (mPlayer.selItem + 1) % ITEM_COUNT;
        if (mPlayer.selItem == ITEM_SCANNER && mPlayer.scannerUses == 0)
            mPlayer.selItem = (mPlayer.selItem + 1) % ITEM_COUNT;

        if (mPlayer.selItem == ITEM_SCANNER)
        {
            mKeyLeft = mKeyRight = mKeyUp = mKeyDown = false;
            mPlayer.dir = DIR_DOWN;
            //mPlayer.state = PLR_SCANNING;
        }
    }

    if (event.Key == KEY_KEY_C)
    {
        srand(mStateTime);
        int pxi = (int)mPlayer.x, pyi = (int)mPlayer.y;
        if (pxi >= 0 && pxi < mLevelWidth && pyi >= 0 && pyi < mLevelHeight)
        {
            Tile &t = mWorld[pyi * mLevelWidth + pxi];
            if (t.type == TT_SHOP_DRILL && mScore >= DRILL_COST)
            {
                mScore -= DRILL_COST;
                mPlayer.drillUses += rand() % (DRILL_USES_MAX - DRILL_USES_MIN + 1) + DRILL_USES_MIN;
                FSOUND_PlaySound(FSOUND_FREE, sSamples[SAMPLE_DRILL1]);
                mPlayer.selItem = ITEM_DRILL;
            }
            if (t.type == TT_SHOP_SCANBAT && mScore >= SCANBAT_COST)
            {
                mScore -= SCANBAT_COST;
                ++mPlayer.scannerUses;
                FSOUND_PlaySound(FSOUND_FREE, sSamples[SAMPLE_SCANNER]);
            }

            
        }
    }

    if (event.Key == KEY_KEY_Z && mPlayer.scannerUses > 0 && mPlayer.state == PLR_MOVING && mPlayer.selItem == ITEM_SCANNER)
    {
        mScanResult = 0;
        mScanResultTransition = 0;
        for (int ty = 0; ty < mLevelHeight; ++ty)
        {
            for (int tx = 0; tx < mLevelWidth; ++tx)
            {
                Tile &t = mWorld[ty * mLevelWidth + tx];
                t.scannerInfluence = false;
                float tmx = tx + 0.5f;
                float tmy = ty + 0.5f;
                float px = mPlayer.x;
                float py = mPlayer.y;
                float scanWidth = (1.0f - mPlayer.scannerDist) * (SCANNER_MAX_ANGLE - SCANNER_MIN_ANGLE) + SCANNER_MIN_ANGLE;
                float scanRange = mPlayer.scannerDist * (SCANNER_MAX_DIST - SCANNER_MIN_DIST) + SCANNER_MIN_DIST;
                float p_to_t_angle = atan2f(tmy - py, tmx - px);
                float scanAngleToTDelta = unwrap(mPlayer.scannerAngle, p_to_t_angle, (float)(2 * PI));
                //if (fabsf(mPlayer.scannerAngle - scanAngleToTDelta) <= scanWidth / 2) ...
                bool withinAngle = angleBetween(p_to_t_angle, mPlayer.scannerAngle - scanWidth / 2, mPlayer.scannerAngle + scanWidth / 2);
                bool withinRange = ((px - tmx) * (px - tmx) + (py - tmy) * (py - tmy)) <= (scanRange * scanRange);
                if (withinAngle && withinRange)
                {
                    t.scannerInfluence = true;
                    if (t.hasGold)
                        ++mScanResult;
                }
            }
        }
        --mPlayer.scannerUses;
        mPlayer.state = PLR_SCANNING;
        mPlayer.stateTime = 0;
        FSOUND_PlaySound(FSOUND_FREE, sSamples[SAMPLE_SCANNER]);
    }

    return false;
}

bool MainState::onMouseDown(const irr::SEvent::SMouseInput & /*event*/)
{
    if (mTitleUIActive || mEndingUIActive)
        mIMGUI.onMouseDown();

    return false;
}

bool MainState::onMouseUp(const irr::SEvent::SMouseInput & /*event*/)
{
    mIMGUI.onMouseUp();
    return false;
}

bool MainState::onMouseMotion(const irr::SEvent::SMouseInput &event)
{
    mLastMousePos.set(event.X, event.Y);
    mIMGUI.onMouseMotion(event.X, event.Y);
    return false;
}

bool MainState::onMouseWheel(const irr::SEvent::SMouseInput & /*event*/)
{
    return false;
}


bool MainState::update(I32 deltaTime)
{
    bool result = AppState::update(deltaTime);
    if (!result)
        return false; // game is paused...

    if (mTitleUIActive)
    {
        return false;
    }

    mScanResultTransition += 0.005f;

    /*
    if (mInGame && game_should_end)
    {
        mInGame = false;
        mEndingUIActive = true;
        //FSOUND_PlaySound(FSOUND_FREE, sSamples[SAMPLE_END]);
    }
    */

    // Following is executed only when game is not paused & game is running.

    float plrOX = mPlayer.x, plrOY = mPlayer.y;
    mPlayer.stateTime += deltaTime;

    switch (mPlayer.state)
    {
    case PLR_MOVING:
        if (mPlayer.selItem == ITEM_SCANNER)
        {
            if (mKeyUp)    { mPlayer.scannerDist += SCANNER_DIST_CHANGE; }
            if (mKeyDown)  { mPlayer.scannerDist -= SCANNER_DIST_CHANGE; }
            if (mKeyLeft)  { mPlayer.scannerAngle -= SCANNER_ANGLE_CHANGE; }
            if (mKeyRight) { mPlayer.scannerAngle += SCANNER_ANGLE_CHANGE; }
            mPlayer.scannerDist = core::clamp<float>(mPlayer.scannerDist, 0, 1);
            mPlayer.scannerAngle = fmodf(fmodf(mPlayer.scannerAngle, (float)(2 * PI)) + (float)(2 * PI), (float)(2 * PI));
        }
        else
        {
            if (mKeyUp)    { mPlayer.y -= PLR_SPEED; mPlayer.dir = DIR_UP; }
            if (mKeyDown)  { mPlayer.y += PLR_SPEED; mPlayer.dir = DIR_DOWN; }
            if (mKeyLeft)  { mPlayer.x -= PLR_SPEED; mPlayer.dir = DIR_LEFT; }
            if (mKeyRight) { mPlayer.x += PLR_SPEED; mPlayer.dir = DIR_RIGHT; }
            if (mKeyUp || mKeyDown || mKeyLeft || mKeyRight)
                mPlayer.walkTime += deltaTime;
        }
        break;
    case PLR_DIGGING_PICKAXE:
        if (mPlayer.stateTime >= DIG_PICKAXE_TIME)
        {
            mPlayer.state = PLR_MOVING;
            mPlayer.stateTime = 0;
            mPlayer.walkTime = ((mPlayer.walkTime / TIME_PER_FRAME) % 2 + 1) * TIME_PER_FRAME;
        }
        break;
    case PLR_DIGGING_DRILL:
        if (mPlayer.stateTime >= DIG_DRILL_TIME)
        {
            mPlayer.state = PLR_MOVING;
            mPlayer.stateTime = 0;
        }
        break;
    case PLR_SCANNING:
        if (mPlayer.stateTime >= SCANNER_TIME)
        {
            mPlayer.state = PLR_MOVING;
            mPlayer.stateTime = 0;
        }
        break;
    }

    for (int ty = 0; ty < mLevelHeight; ++ty)
    {
        for (int tx = 0; tx < mLevelWidth; ++tx)
        {
            Tile &t = mWorld[ty * mLevelWidth + tx];
            if (t.type == TT_GROUND && t.conditionTarget < t.condition)
            {
                t.condition -= GROUND_KILL_SPEED;
                if (t.condition < t.conditionTarget)
                    t.condition = t.conditionTarget;
                if (t.condition <= 0)
                {
                    t.condition = 1.0f;
                    t.type = TT_TUNNEL;
                }
            }
        }
    }

    core::rectf plrArea(mPlayer.x - PLR_W / 2, mPlayer.y - PLR_H / 2, mPlayer.x + PLR_W / 2, mPlayer.y + PLR_H / 2);
    bool plrColl = false;
    for (int ty = (int)mPlayer.y - 1; ty <= (int)mPlayer.y + 1 && !plrColl; ++ty)
    {
        if (ty < 0 || ty >= mLevelHeight)
            continue;
        for (int tx = (int)mPlayer.x - 1; tx <= (int)mPlayer.x + 1 && !plrColl; ++tx)
        {
            if (tx < 0 || tx >= mLevelWidth)
                continue;
            Tile &t = mWorld[ty * mLevelWidth + tx];
            core::rectf tileArea(tx + 0.5f - t.condition * 0.5f, ty + 0.5f - t.condition * 0.5f, tx + 0.5f + t.condition * 0.5f, ty + 0.5f + t.condition * 0.5f);
            if (t.type == TT_GROUND)
            {
                if (t.conditionTarget >= GROUND_KILL_THRESHOLD)
                {
                    if (tileArea.isRectCollided(plrArea))
                    {
                        mPlayer.x = plrOX;
                        mPlayer.y = plrOY;
                        if (mPlayer.selItem == ITEM_DRILL && mPlayer.drillUses > 0)
                        {
                            mPlayer.state = PLR_DIGGING_DRILL;
                            --mPlayer.drillUses;
                            t.conditionTarget -= DIG_DRILL_HIT;
                            FSOUND_PlaySound(FSOUND_FREE, sSamples[SAMPLE_DRILL_FIRST + rand() % SAMPLE_DRILL_COUNT]);
                            if (mPlayer.drillUses <= 0)
                            {
                                mPlayer.selItem = ITEM_PICKAXE;
                                FSOUND_PlaySound(FSOUND_FREE, sSamples[SAMPLE_PICKAXE5]);
                            }
                        }
                        else
                        {
                            mPlayer.state = PLR_DIGGING_PICKAXE;
                            t.conditionTarget -= DIG_PICKAXE_HIT;
                            FSOUND_PlaySound(FSOUND_FREE, sSamples[SAMPLE_PICKAXE_FIRST + rand() % SAMPLE_PICKAXE_COUNT]);
                            mScore -= PICKAXE_USE_COST;
                        }
                        if (t.conditionTarget <= GROUND_KILL_THRESHOLD)
                        {
                            t.conditionTarget = 0;
                            if (t.hasGold)
                            {
                                t.hasGold = false;
                                mScore += DIAMOND_GOLD_VALUE;
                                FSOUND_PlaySound(FSOUND_FREE, sSamples[SAMPLE_COLLECT]);
                                ++mPlayer.diamondsFound;
                                if (mPlayer.diamondsFound >= GOAL_DIAMONDS)
                                {
                                    mEndingUIActive = true;
                                    mInGame = false;
                                }
                            }
                        }
                        mPlayer.stateTime = 0;
                        
                        plrColl = true;
                        if (t.conditionTarget <= 0)
                            FSOUND_PlaySound(FSOUND_FREE, sSamples[SAMPLE_EXPL_FIRST + rand() % SAMPLE_EXPL_COUNT]);
                    }
                }
            }
            else if (t.type == TT_IMPASSABLE)
            {
                if (tileArea.isRectCollided(plrArea))
                {
                    mPlayer.x = plrOX;
                    mPlayer.y = plrOY;
                }
            }
        }
    }

    return true;
}


void MainState::render(irr::IrrlichtDevice *device)
{
    assert(device);
    irr::video::IVideoDriver *driver = device->getVideoDriver();
    assert(driver);

    driver->beginScene(true, true, video::SColor(0x00445566));

    Graphics *g = App::getInstance()->getGraphics();
    g->beginRender();

    if (mInGame)
    {
        renderGame(device);
    }
    else
    {
        mIMGUI.begin(mStateTime);

        if (mTitleUIActive)
        {
            mFont->drawString(WIDTH / 2, 150, "Drill Warrior", IF_ALIGN_HORIZ_CENTER, 1, Color(0xffffffff));
            mFont->drawString(WIDTH / 2, 180, "Game for Ludum Dare 19 - Created in less than 2 days", IF_ALIGN_HORIZ_CENTER, 0.5f, Color(0x80ffffff));
            mFont->drawString(WIDTH / 2, 210, gVersionString, IF_ALIGN_HORIZ_CENTER, 0.4f, Color(0x60ccccff));

            mFont->drawText(WIDTH / 2, 250,
                            "There are 5 diamonds in the ground, each worth a fortune.\n"
                            "Your task is to dig them all up. But it's not cheap to dig\n"
                            "around. To help with that, you can use a scanner to search\n"
                            "for the diamonds, but even the scanner batteries aren't free.\n"
                            "So use your cash wisely. Good luck!",
                            IF_ALIGN_HORIZ_CENTER, 0.5f, 0, Color(0xaaffffff));

            if (mIMGUI.button(WIDTH / 2 - mButtonStyle->getWidth() / 2, 350, IMGUI_AI_COUNTER, mButtonStyle, 0, 0, "Play"))
            {
                startGame();
                mTitleUIActive = false;
                mInGame = true;
                mGameLogicTime = 0;
                FSOUND_PlaySound(FSOUND_FREE, sSamples[SAMPLE_COLLECT]);
            }
            if (mIMGUI.button(WIDTH / 2 - mButtonStyle->getWidth() / 2, 450, IMGUI_AI_COUNTER, mButtonStyle, 0, 0, "Quit"))
                App::getInstance()->setAlive(false);

            mFont->drawString(WIDTH / 2, HEIGHT - 5, "Copyright 2010 Jetro Lauha - www.jet.ro", IF_ALIGN_BOTTOM_HCENTER, 0.5f, Color(0x80ffffff));
        }
        else if (mEndingUIActive)
        {
            char *title = "You found all diamonds!";
            char buf[100];
            sprintf(buf, "%d", mScore);
            if (mScore < 0)
                title = "Failboat!";
            mFont->drawString(WIDTH / 2, 200, title, IF_ALIGN_HORIZ_CENTER, 1, Color(0xffffeedd));
            mFont->drawString(WIDTH / 2, 300, "Final score:", IF_ALIGN_HORIZ_CENTER, 0.6f, Color(0x80ffccaa));
            mFont->drawString(WIDTH / 2, 350, buf, IF_ALIGN_HORIZ_CENTER, 0.8f, Color(0xccffccaa));
            if (mIMGUI.button(WIDTH / 2 - mButtonStyle->getWidth() / 2, 450, IMGUI_AI_COUNTER, mButtonStyle, 0, 0, "Quit"))
                App::getInstance()->setAlive(false);
        }

        mIMGUI.end();
    }

    g->endRender();

    driver->endScene();
}


void MainState::genCone(float x, float y, float z, float r, const video::SColor &color)
{
    if (mVertices.size() >= 65000)
    {
        assert(0 && "too much triangles");
        return;
    }
    const float centerU = 0.5f + (rand() % 255) / 256.f + 2;
    const float centerV = 0.5f + (rand() % 255) / 256.f + 2;
    const float uvW = 0.5f, uvH = 0.5f;
    const float MAX_R = 128;
    const float Z_NEAR = 0.1f, Z_FAR = 1.1f;//MIN(r / MAX_R, 1.0f);
    const int SEGMENTS = 15;
    const float SEG_RAD = (float)(2 * PI / SEGMENTS);
    const int preVtxCount = (int)mVertices.size();
    static core::position2df unitCircleSegmentVertices[SEGMENTS];
    static bool unitCircleSegmentVerticesInitialized = false;
    if (!unitCircleSegmentVerticesInitialized)
    {
        float angle = 0;
        for (int a = 0; a < SEGMENTS; ++a, angle += SEG_RAD)
            unitCircleSegmentVertices[a].set(cosf(angle), sinf(angle));
        unitCircleSegmentVerticesInitialized = true;
    }
    int idx = preVtxCount;
    video::SColorf dimmedColor(color);
    dimmedColor.r *= 0.9f;
    dimmedColor.g *= 0.9f;
    dimmedColor.b *= 0.9f;
    S3DVertex vtx(x, y, Z_NEAR + z, 0, 0, Z_NEAR + z, dimmedColor.toSColor(), centerU, centerV);
    mVertices.push_back(vtx);
    vtx.Color = color;
    for (int seg = 0; seg < SEGMENTS; ++seg)
    {
        float sr = r * ((85 + (rand() & 31)) / 100.f);
        vtx.Pos.set(x + unitCircleSegmentVertices[seg].X * sr, y + unitCircleSegmentVertices[seg].Y * sr, Z_FAR + z);
        vtx.TCoords.set(unitCircleSegmentVertices[seg].X * 0.5f + centerU, unitCircleSegmentVertices[seg].Y * 0.5f + centerV);
        vtx.Normal.set(unitCircleSegmentVertices[seg].X, unitCircleSegmentVertices[seg].Y, 0);
        mVertices.push_back(vtx);
    }
    for (int seg = 0; seg < SEGMENTS; ++seg)
    {
        mIndices.push_back((u16)idx);   // 1st vertex for each tri is the center one
        mIndices.push_back((u16)idx + 1 + seg);
        mIndices.push_back((u16)idx + 1 + (seg + 1) % SEGMENTS);
    }
}


void MainState::genArc(float x, float y, float r, float uvr, float a1, float a2, const video::SColor &color)
{
    if (mVertices.size() >= 65000)
    {
        assert(0 && "too much triangles");
        return;
    }
    const float uvScale = uvr;
    const float centerU = 0.5f + cosf(mStateTime * 0.001f) * 0.01f;
    const float centerV = 0.5f + sinf(mStateTime * 0.001f) * 0.01f;
    const float uvW = 0.5f, uvH = 0.5f;
    const float MAX_R = 128;
    const int SEGMENTS = 12;
    const float SEG_RAD = (float)(2 * PI / SEGMENTS);
    const int preVtxCount = (int)mVertices.size();
    int idx = preVtxCount;
    video::SColorf dimmedColor(color);
    dimmedColor.a *= 0.5f;
    S3DVertex vtx(x, y, 0, 0, 0, 0, dimmedColor.toSColor(), centerU, centerV);
    mVertices.push_back(vtx);
    vtx.Color = color;
    for (int seg = 0; seg <= SEGMENTS; ++seg)
    {
        float ang = a1 + seg * (a2 - a1) / SEGMENTS;
        float ca = cosf(ang), sa = sinf(ang);
        vtx.Pos.set(x + ca * r, y + sa * r, 0);
        vtx.TCoords.set(ca * uvScale + centerU + cosf((mStateTime + seg * 10) * 0.0013f) * 0.01f, sa * uvScale + centerV + sinf((mStateTime + seg * 10) * 0.0014f) * 0.01f);
        vtx.Normal.set(ca, sa, 0);
        mVertices.push_back(vtx);
    }
    for (int seg = 0; seg < SEGMENTS; ++seg)
    {
        mIndices.push_back((u16)idx);   // 1st vertex for each tri is the center one
        mIndices.push_back((u16)idx + 1 + seg);
        mIndices.push_back((u16)idx + 1 + seg + 1);
    }
}


void MainState::drawSprite(float x, float y, int num, bool horizFlip)
{
    static Image *spriteImg = mImages[IMG_SPRITES];
    static const int spriteSize = 64;
    static const int spritesPerRow = (int)(spriteImg->getWidth() / spriteSize);
    int col = num % spritesPerRow;
    int row = num / spritesPerRow;
    Rectangle srcRect((float)(col * spriteSize), (float)(row * spriteSize), (float)spriteSize, (float)spriteSize);
    spriteImg->draw(IDM_NORMAL, IF_ALIGN_CENTER | (horizFlip ? IF_FLIP_HORIZ : 0), x, y, Color(0xffffffff), &srcRect);
}


static float sClamp(float value, float min, float max)
{
    return value < min ? min : (value > max ? max : value);
}

static float sStep(float value, float stepPosition)
{
    return (float)(value < stepPosition ? 0 : 1);
}

static float sBoxStep(float value, float slopeStart, float slopeEnd)
{
    float diff = slopeEnd - slopeStart;
    if (diff == 0)
        return sStep(value, slopeStart);

    return sClamp((value - slopeStart) / diff, 0, 1);
}

static float sBoxPulse(float value, float upSlopeStart, float upSlopeEnd,
                       float downSlopeStart, float downSlopeEnd)
{
    return sBoxStep(value, upSlopeStart, upSlopeEnd) -
           sBoxStep(value, downSlopeStart, downSlopeEnd);
}


void MainState::renderGame(irr::IrrlichtDevice *device)
{
    irr::video::IVideoDriver *driver = device->getVideoDriver();
    App *app = App::getInstance();
    Graphics *g = app->getGraphics();

    //mPlayer.x += 0.02f;
    //mPlayer.y += 0.1f;

    const float scale = 64.f;
    const float scx = app->getScreenWidth() / 2.0f, scy = app->getScreenHeight() / 2.0f;
    float cx = mPlayer.x, cy = mPlayer.y;
    int tileVisTop = core::clamp<int>((int)(cy - scy / scale - 1), 0, mLevelHeight - 1);
    int tileVisBottom = core::clamp<int>((int)(cy + scy / scale + 1), 0, mLevelHeight - 1);
    int tileVisLeft = core::clamp<int>((int)(cx - scx / scale - 1), 0, mLevelWidth - 1);
    int tileVisRight = core::clamp<int>((int)(cx + scx / scale + 1), 0, mLevelWidth - 1);
    mVertices.set_used(0);
    mIndices.set_used(0);

    video::SMaterial mat;
    mat.MaterialType = video::EMT_ONETEXTURE_BLEND;
    mat.MaterialTypeParam = video::pack_texureBlendFunc(video::EBF_SRC_ALPHA, video::EBF_ONE_MINUS_SRC_ALPHA);
    mat.setFlag(video::EMF_LIGHTING, false);
    mat.setTexture(0, ((ImageIrr *)mImages[IMG_CAVE])->getIrrTexture());
    mat.TextureLayer[0].TextureWrapU = mat.TextureLayer[0].TextureWrapV = video::ETC_REPEAT;

    // draw level
    float scannerInfluence = 0;
    if (mPlayer.state == PLR_SCANNING)
        scannerInfluence = sBoxPulse((float)mPlayer.stateTime, 0.f, SCANNER_TIME / 2.f, SCANNER_TIME / 2.f, (float)SCANNER_TIME);
    float drillShopX = 0, drillShopY = 0;
    float scanShopX = 0, scanShopY = 0;
    bool drillShopVis = false, scanShopVis = false;
    for (int ty = tileVisTop; ty <= tileVisBottom; ++ty)
    {
        for (int tx = tileVisLeft; tx <= tileVisRight; ++tx)
        {
            srand(ty * 1001 + tx);
            Tile &t = mWorld[ty * mLevelWidth + tx];
            float x = tx + t.x - cx + (rand() % 21 - 10) / 80.f;
            float y = ty + t.y - cy + (rand() % 21 - 10) / 80.f;
            //core::dimension2du rs((int)scale, (int)scale);
            static const float r = sqrtf((scale / 2) * (scale / 2) + (scale / 2) * (scale / 2)) + 1;
            switch (t.type)
            {
            case TT_OPEN_AIR:
                //driver->draw2DRectangle(video::SColor(0xff88eeff), core::recti(core::vector2di((int)(x * scale - scale / 2 + scx), (int)(y * scale - scale / 2 + scy)), rs));
                genCone(x * scale + scx, y * scale + scy, 1, r + 10, video::SColor(0xff88ee99));
                break;
            case TT_IMPASSABLE:
                //driver->draw2DRectangle(video::SColor(0xff444444), core::recti(core::vector2di((int)(x * scale - scale / 2 + scx), (int)(y * scale - scale / 2 + scy)), rs));
                genCone(x * scale + scx, y * scale + scy, 0, r, video::SColor(0xff444444));
                break;
            case TT_SHOP_DRILL:
                genCone(x * scale + scx, y * scale + scy, -0.1f, r * 1.15f, video::SColor(0xffaa8855));
                drillShopX = x * scale + scx;
                drillShopY = y * scale + scy;
                drillShopVis = true;
                break;
            case TT_SHOP_SCANBAT:
                genCone(x * scale + scx, y * scale + scy, -0.1f, r * 1.15f, video::SColor(0xffaa8855));
                scanShopX = x * scale + scx;
                scanShopY = y * scale + scy;
                scanShopVis = true;
                break;
            case TT_GROUND:
                {
                    //tunnel bkg
                    //driver->draw2DRectangle(video::SColor(0xff775533), core::recti(core::vector2di((int)(x * scale - scale / 2 + scx), (int)(y * scale - scale / 2 + scy)), rs));
                    genCone(x * scale + scx, y * scale + scy, 1, r, video::SColor(0xff775533));
                    //core::dimension2du rsg = rs;
                    //rsg.Width = rsg.Height = (int)(t.condition * scale);
                    //driver->draw2DRectangle(video::SColor(0xff997766), core::recti(core::vector2di((int)(x * scale - rsg.Width / 2 + scx), (int)(y * scale - rsg.Height / 2 + scy)), rsg));
                    genCone(x * scale + scx, y * scale + scy, 0, r * (t.condition - (t.scannerInfluence ? 0.25f : 0.0f) * scannerInfluence), video::SColor(0xff997766));
                    if (t.hasGold)
                    {
                        genCone(x * scale + scx, y * scale + scy, 0.25f * t.condition, r * t.condition + scale * 0.02f, video::SColor(0xffffddff));
                    }
                }
                break;
            case TT_TUNNEL:
                //driver->draw2DRectangle(video::SColor(0xff775533), core::recti(core::vector2di((int)(x * scale - scale / 2 + scx), (int)(y * scale - scale / 2 + scy)), rs));
                genCone(x * scale + scx, y * scale + scy, 1, r, video::SColor(0xff775533));
                break;
            }
        }
    }

    g->endRender();
    driver->setMaterial(mat);
    driver->drawVertexPrimitiveList(mVertices.const_pointer(), mVertices.size(),
                                    mIndices.const_pointer(), mIndices.size() / 3,
                                    video::EVT_STANDARD, scene::EPT_TRIANGLES, video::EIT_16BIT);
    g->beginRender();

    if (drillShopVis)
    {
        int pxi = (int)mPlayer.x, pyi = (int)mPlayer.y;
        drawSprite(drillShopX + 10, drillShopY - 10, SPR_DRILL);
        if (pxi >= 0 && pxi < mLevelWidth && pyi >= 0 && pyi < mLevelHeight)
        {
            if (mWorld[pyi * mLevelWidth + pxi].type == TT_SHOP_DRILL && mScore >= DRILL_COST)
                mFont->drawText(drillShopX, drillShopY - scale * 0.5f, "Press C\nto buy", IF_ALIGN_CENTER, 0.5f);
            else
                mFont->drawText(drillShopX, drillShopY - scale * 0.5f, "Drill\nShop", IF_ALIGN_CENTER, 0.5f);
        }
        mFont->drawText(drillShopX, drillShopY + scale * 0.5f, DRILL_COST_STR, IF_ALIGN_CENTER, 0.5f, 0, Color(0xff000000));
    }
    if (scanShopVis)
    {
        int pxi = (int)mPlayer.x, pyi = (int)mPlayer.y;
        drawSprite(scanShopX, scanShopY, SPR_SCANNER);
        if (pxi >= 0 && pxi < mLevelWidth && pyi >= 0 && pyi < mLevelHeight)
        {
            if (mWorld[pyi * mLevelWidth + pxi].type == TT_SHOP_SCANBAT && mScore >= SCANBAT_COST)
                mFont->drawText(scanShopX, scanShopY - scale * 0.5f, "Press C\nto buy", IF_ALIGN_CENTER, 0.5f);
            else
                mFont->drawText(scanShopX, scanShopY - scale * 0.5f, "Scanner\nBatteries", IF_ALIGN_CENTER, 0.5f);
        }
        mFont->drawText(scanShopX, scanShopY + scale * 0.5f, SCANBAT_COST_STR, IF_ALIGN_CENTER, 0.5f, 0, Color(0xff000000));
    }

    srand(mStateTime);
    // draw player
    {
        //float px = (mPlayer.x - cx) * scale - scale * PLR_W * 0.5f + scx;
        //float py = (mPlayer.y - cy) * scale - scale * PLR_H * 0.5f + scy;
        float px = (mPlayer.x - cx) * scale + scx;
        float py = (mPlayer.y - cy) * scale + scy;
        float opx = px, opy = py;
        if (mPlayer.state == PLR_DIGGING_PICKAXE)
            py -= scale * PLR_H / 4 * (fabsf((float)cos(mPlayer.stateTime * PI / DIG_PICKAXE_TIME + PI / 2)));
        else if (mPlayer.state == PLR_DIGGING_DRILL)
        {
            px += (rand() % 21 - 10) / 10.f;
            py += (rand() % 21 - 10) / 10.f;
        }
        //driver->draw2DRectangle(video::SColor(0xff88ccaa), core::recti(core::vector2di((int)px, (int)py), core::dimension2du((int)(scale * PLR_W), (int)(scale * PLR_H))));
        if (mPlayer.state == PLR_DIGGING_PICKAXE && mPlayer.dir != DIR_DOWN)
            drawSprite(opx, opy - 8, SPR_PICKAXE, mPlayer.dir != DIR_LEFT);
        if (mPlayer.state == PLR_DIGGING_DRILL && mPlayer.dir != DIR_DOWN)
            drawSprite(opx, opy - 8, SPR_DRILL, mPlayer.dir != DIR_LEFT);
        drawSprite(px, py, mPlayer.dir * FRAMES_PER_DIR + (mPlayer.walkTime / TIME_PER_FRAME) % FRAMES_PER_DIR);
        if (mPlayer.state == PLR_DIGGING_PICKAXE && mPlayer.dir == DIR_DOWN)
            drawSprite(opx, opy, SPR_PICKAXE);
        if (mPlayer.state == PLR_DIGGING_DRILL && mPlayer.dir == DIR_DOWN)
            drawSprite(opx, opy, SPR_DRILL);
    }

    // draw scanner
    if (mPlayer.state == PLR_SCANNING || (mPlayer.state == PLR_MOVING && mPlayer.selItem == ITEM_SCANNER && mPlayer.scannerUses > 0))
    {
        mVertices.set_used(0);
        mIndices.set_used(0);
        float px = (mPlayer.x - cx) * scale + scx;
        float py = (mPlayer.y - cy) * scale + scy;
        float scanWidth = (1.0f - mPlayer.scannerDist) * (SCANNER_MAX_ANGLE - SCANNER_MIN_ANGLE) + SCANNER_MIN_ANGLE;
        float scanRange = mPlayer.scannerDist * (SCANNER_MAX_DIST - SCANNER_MIN_DIST) + SCANNER_MIN_DIST;
        genArc(px, py, scanRange * scale, mPlayer.scannerDist * 4 + 2.51f, mPlayer.scannerAngle - scanWidth / 2, mPlayer.scannerAngle + scanWidth / 2, video::SColor(0xff00ff00));
        drawSprite(px, py, SPR_SCANNER);
        g->endRender();
        mat.setFlag(video::EMF_ZBUFFER, false);
        mat.setTexture(0, ((ImageIrr *)mImages[IMG_SCANNER])->getIrrTexture());
        driver->setMaterial(mat);
        driver->drawVertexPrimitiveList(mVertices.const_pointer(), mVertices.size(),
                                        mIndices.const_pointer(), mIndices.size() / 3,
                                        video::EVT_STANDARD, scene::EPT_TRIANGLES, video::EIT_16BIT);
        g->beginRender();
    }

    char buf[100];

    if (mScanResultTransition < 1.0f)
    {
        sprintf(buf, "Last scanner result:\n%d found", mScanResult);
        mFont->drawText(300, app->getScreenHeight() - 50, buf, IF_ALIGN_TOP_HCENTER, 0.6f, 0, Color(1, 1, 1, 1.0f - mScanResultTransition));
    }


    //sprintf(buf, "plr.state %d", mPlayer.state);
    //mFont->drawString(200, 0, buf, IF_ALIGN_TOP_LEFT, 0.5f, turska::Color(0xff000000));
    //sprintf(buf, "plr.stateTime %d", mPlayer.stateTime);
    //mFont->drawString(200, 20, buf, IF_ALIGN_TOP_LEFT, 0.5f, turska::Color(0xff000000));

    // draw score
    sprintf(buf, "$%d", mScore);
    mFont->drawString(39, 39, buf, IF_ALIGN_TOP_LEFT, 1, Color(0x88880000));
    mFont->drawString(41, 42, buf, IF_ALIGN_TOP_LEFT, 1, Color(0x99aa4400));
    mFont->drawString(40, 40, buf, IF_ALIGN_TOP_LEFT, 1, Color(0xffffee00));
    mFont->drawString(40, 65, "assets", IF_ALIGN_TOP_LEFT, 0.5f, Color(0x55000000));

    // draw items
    {
        float y = app->getScreenHeight() - (ITEM_COUNT + 1) * mFont->getFontHeight(0.5f);
        for (int a = 0; a < ITEM_COUNT; ++a)
        {
            if (a == ITEM_SCANNER)
                sprintf(buf, "%s %d", sItemNames[a], mPlayer.scannerUses);
            else
                strcpy(buf, sItemNames[a]);
            Color c(mPlayer.selItem == a ? 0xccffffff : 0x8a000000);
            if (a == ITEM_DRILL && mPlayer.drillUses <= 0)
                c.a /= 2;
            mFont->drawString(100, y + a * mFont->getFontHeight(0.5f), buf, IF_ALIGN_TOP_LEFT, 0.5f, c);
            if (mPlayer.selItem == a)
                mFont->drawString(95, y + a * mFont->getFontHeight(0.5f), ">", IF_ALIGN_TOP_RIGHT, 0.5f, Color(0xffffffff));
        }
        if (mPlayer.selItem == ITEM_SCANNER && mPlayer.scannerUses > 0)
        {
            mFont->drawString(190, y + mFont->getFontHeight(0.25f), "Press Z", IF_ALIGN_TOP_HCENTER, 0.5f, turska::Color(0xff444444));
            mFont->drawString(190, y + mFont->getFontHeight(0.75f), "to use", IF_ALIGN_TOP_HCENTER, 0.5f, turska::Color(0xff444444));
        }
        y += mFont->getFontHeight(0.25f);
        mFont->drawString(40, y, "Press X", IF_ALIGN_TOP_HCENTER, 0.5f, turska::Color(0xffbbbbbb));
        y += mFont->getFontHeight(0.5f);
        mFont->drawString(40, y, "to change", IF_ALIGN_TOP_HCENTER, 0.5f, turska::Color(0xffbbbbbb));
    }
}
